# output readers
from pypact.reader import InventoryReader as Reader
from pypact.reader import JSONReader
from pypact.filerecord import InventoryFileRecord
from pypact.runner import compute

# output data structures
from pypact.output.rundata import RunData
from pypact.output.doserate import DoseRate
from pypact.output.nuclide import Nuclide
from pypact.output.nuclides import Nuclides
from pypact.output.output import Output
from pypact.output.gammaspectrum import GammaSpectrum
from pypact.output.timestep import TimeStep

# printlib data structures
from pypact.printlib.printlib5 import PrintLib5, PrintLib5FileRecord, PrintLib5Reader
from pypact.printlib.printlib4 import PrintLib4, PrintLib4FileRecord, PrintLib4Reader

# input
from pypact.input.fispactinput import FispactInput
from pypact.input.fluxesfile import FluxesFile, ArbFluxesFile
from pypact.input.filesfile import FilesFile
from pypact.input.inputdata import InputData
from pypact.input.serialization import to_file, from_file, to_string
from pypact.input.groupstructures import ALL_GROUPS
from pypact.input.keywords import CONTROL_KEYWORDS, \
    INIT_KEYWORDS, \
    INVENTORY_KEYWORDS, \
    OVER_SUBKEYWORDS, \
    DEPRECATED_KEYWORDS

# projectiles
from pypact.library.projectiles import PROJECTILE_NEUTRON, \
    PROJECTILE_DEUTERON, \
    PROJECTILE_PROTON, \
    PROJECTILE_ALPHA, \
    PROJECTILE_GAMMA, \
    get_projectile_name, \
    get_projectile_symbol, \
    get_projectile_value, \
    VALID_PROJECTILE_NAMES, \
    VALID_PROJECTILE_SYMBOLS, \
    VALID_PROJECTILES

# library
from pypact.library.nuclidelib import NUCLIDE_DICTIONARY, \
    NUMBER_OF_ELEMENTS, \
    NUMBER_OF_ISOTOPES, \
    find_isotopes, \
    find_element, \
    find_z, \
    get_all_isotopes

from pypact.library.reactionlib import REACTION_DICTIONARY, \
    getreaction, \
    getmt

# utilities
from pypact.util.exceptions import PypactException
from pypact.util.loglevels import *


# This makes importing slow, keep it seperate
#import pypact.analysis



from pypact.util.decorators import freeze_it
from pypact.util.jsonserializable import JSONSerializable
from pypact.util.lines import line_indices
from pypact.util.exceptions import PypactNotPrintLib5FileException
from pypact.util.numerical import get_float
from pypact.printlib.tags import PRINTLIB5_HEADER
from pypact.filerecord import FileRecord
from pypact.reader import Reader


@freeze_it
class PrintLib5FileRecord(FileRecord):

    def _setup(self):
        indx = line_indices(self.cachedlines, PRINTLIB5_HEADER)
        assert len(indx) == 2
        
        # remove title and next empty line
        self.mean_start_index = indx[0]+ 2
        self.line_start_index = indx[1]
        self.nr_of_entries = 0

    def _process(self):
        if not line_indices(self.cachedlines, "Nuclide Nuclide Nuclide"):
            raise PypactNotPrintLib5FileException(
                "Not a valid printlib5 file or SAVELINES was not used, as no spectral data exists.")

        # off set start index to remove header (column titles are 2 lines)
        self.mean_start_index = self.mean_start_index + 2
        self.nr_of_entries = self.line_start_index - self.mean_start_index

@freeze_it
class SpectralLineData(JSONSerializable):
    """
        Spectral line data
    """
    def __init__(self):    
        self.energies = []
        self.energies_unc = []
        self.intensities = []
        self.intensities_unc = []
        self.norms = []
        self.norms_unc = []

    def __len__(self):
        return len(self.energies)

    def __getitem__(self, index):
        return (self.energies[index],
                self.energies_unc[index],
                self.intensities[index],
                self.intensities_unc[index],
                self.norms[index],
                self.norms_unc[index])

    def addline(self, energy, energy_unc, intensity, intensity_unc, norm, norm_unc):
        self.energies.append(energy)
        self.energies_unc.append(energy_unc)
        self.intensities.append(intensity)
        self.intensities_unc.append(intensity_unc)
        self.norms.append(norm)
        self.norms_unc.append(norm_unc)

@freeze_it
class SpectralData(JSONSerializable):
    """
        Spectral data
    """
    def __init__(self):
        self.name = ""
        self.zai = 0
        self.number = 0
        self.type = ""
        self.nr_of_lines = 0
        self.mean_energy = 0.0
        self.mean_energy_unc = 0.0
        self.mean_normalisation = 0.0
        self.mean_normalisation_unc = 0.0
        self.lines = SpectralLineData()

    def fispact_deserialize(self, linedump):
        self.__init__()

        line = linedump[0].strip('\n')

        # fortran format is:
        # format(2x,a6,i8,i6,3x,a9,6x,i8,2x,es13.5,' +-',es12.5,es13.5,' +-',es12.5)
        self.name = line[2:8].strip()
        self.zai = int(line[8:16])
        self.number = int(line[16:22])

        # this is a really horrible hack, but the logic used to make the output file
        # is really bad - "no spectral data" is output if no data exists, but this has
        # length 16 but otherwise the length of type is 9 +6 whitespace = 15, so just 1 off
        # why did they do this???
        # Because of this stupid logic we have to hack here
        self.type = line[25:40].strip()

        # if no spectral data then skip
        if self.type == "no spectral dat":
            self.type = "no spectral data"
            return

        self.nr_of_lines = int(line[40:48])
        self.mean_energy = get_float(line[50:63])
        self.mean_energy_unc = get_float(line[66:78])
        self.mean_normalisation = get_float(line[78:91])
        self.mean_normalisation_unc = get_float(line[94:106])

@freeze_it
class PrintLib5(JSONSerializable):
    """
        An object to represent the Printlib5 output
    """
    def __init__(self):
        self.spectral_data = []
        self.nr_of_zais = 0

    def __len__(self):
        return len(self.spectral_data)

    def __getitem__(self, index):
        return self.spectral_data[index]

    def json_deserialize(self, j, objtype=object):
        super().json_deserialize(j)
        self.json_deserialize_list(j, 'spectral_data', SpectralData)

    def fispact_deserialize(self, filerecord):
        self.__init__()

        # ignore header and column headers
        current_line_index = filerecord.line_start_index + 4

        # printlib file loops over ZAIs then loops over decay types
        for i in range(filerecord.nr_of_entries):
            spectral_data = SpectralData()
            indx = filerecord.mean_start_index+i
            spectral_data.fispact_deserialize(filerecord.cachedlines[indx:indx+1])
            self.nr_of_zais = max(self.nr_of_zais, spectral_data.number)

            # due to a bug in FISPACT-II or nuclear data
            # the number of spectral_data.nr_of_lines can be 0 sometimes
            # with type not equal to "no spectral data"
            # need to account for this 
            if (spectral_data.nr_of_lines == 0) and (spectral_data.type != "no spectral data"):
                self.spectral_data.append(spectral_data)
                continue

            if spectral_data.nr_of_lines > 0:            
                # next find the line data for that ZAI
                # we know that FISPACT-II writes it in the same order as averages
                line_data = filerecord.cachedlines[current_line_index:current_line_index+spectral_data.nr_of_lines]
                current_line_index += spectral_data.nr_of_lines
                # loop over decay types
                for line in line_data:
                    # first 41 characters are redundant - information is same as in mean data section
                    # format after is then 
                    # (3(es13.5,' +-',es12.5))
                    d = line[41:]
                    spectral_data.lines.addline(energy=float(d[:13]),
                                                energy_unc=float(d[16:28]),
                                                intensity=float(d[28:41]),
                                                intensity_unc=float(d[44:56]),
                                                norm=float(d[56:69]),
                                                norm_unc=float(d[72:84]))
                assert spectral_data.nr_of_lines == len(spectral_data.lines)
            else:
                current_line_index += 1

            # append to data
            self.spectral_data.append(spectral_data)

class PrintLib5Reader(Reader):
    """
        It can read fispact printlib 5 file formats
    """
    def __init__(self, filename):
        super().__init__(filename)
        self.record = PrintLib5FileRecord(filename)
        self.output = PrintLib5()

    def __enter__(self):
        self.output.fispact_deserialize(self.record)
        return self.output



import os

from pypact.filerecord import InventoryFileRecord
from pypact.output.output import Output


class Reader(object):
    """
        It can read fispact out file formats
    """

    def __init__(self, filename):
        self.filename = filename

    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass


class InventoryReader(Reader):
    """
        It can read fispact out file formats
    """

    def __init__(self, filename, ignorenuclides=False):
        super().__init__(filename)
        self.record = InventoryFileRecord(filename)
        self.output = Output(ignorenuclides=ignorenuclides)

    def __enter__(self):
        self.output.fispact_deserialize(self.record)
        return self.output


class JSONReader(InventoryReader):
    """
        It can read JSON fispact file formats
    """

    def __enter__(self):
        with open(self.filename, 'rt') as f:
            self.output.json_deserialize(f.read())

        return self.output



PRINTLIB5_HEADER = \
    'D E C A Y    R A D I A T I O N    D I S C R E T E   S P E C T R A'

PRINTLIB4_HEADER = \
    'C O L L A P S E D    C R O S S     S E C T I O N S     I N     B A R N S'

PRINTLIB4_START_HEADER = \
    'parent reaction  daughter cross section   +-error            parent reaction  daughter cross section   +-error'

PRINTLIB4_END_HEADER = \
    'N E U T R O N      S P E C T R U M'


from pypact.util.decorators import freeze_it
from pypact.util.jsonserializable import JSONSerializable
from pypact.util.lines import line_indices
from pypact.util.exceptions import PypactNotPrintLib4FileException
from pypact.util.numerical import get_float
from pypact.printlib.tags import PRINTLIB4_HEADER, \
    PRINTLIB4_START_HEADER, \
    PRINTLIB4_END_HEADER
from pypact.library.reactionlib import getmt
from pypact.filerecord import FileRecord
from pypact.reader import Reader


@freeze_it
class PrintLib4FileRecord(FileRecord):

    def _setup(self):
        indices = line_indices(self.cachedlines, PRINTLIB4_HEADER)
        if len(indices) != 1:
            raise PypactNotPrintLib4FileException(
                "Not a valid printlib4 file.")

        sindx = line_indices(self.cachedlines, PRINTLIB4_START_HEADER)
        if len(sindx) != 1:
            raise PypactNotPrintLib4FileException(
                "Not a valid printlib4 file.")

        eindx = line_indices(self.cachedlines, PRINTLIB4_END_HEADER)
        if len(eindx) != 1:
            raise PypactNotPrintLib4FileException(
                "Not a valid printlib4 file.")
        
        # remove title and next empty line
        self.start_index = sindx[0] + 1
        self.end_index = eindx[0]

@freeze_it
class CollapsedXSData(JSONSerializable):
    # cannot enable slots yet due to json_deserialize_list 
    # does not work without __dict__
    # __slots__ = ['nuclide', 'reaction', 'daughter', 'xs', 'delta_xs']

    def __init__(self, nuclide="", reaction="", 
                daughter="", xs=0.0, delta_xs=0.0):
        self.nuclide = nuclide
        self.reaction = reaction
        self.daughter = daughter
        self.xs = xs
        self.delta_xs = delta_xs

    @property
    def findmt(self):
        return getmt(self.reaction[3:-1].strip())

    def fispact_deserialize(self, line):
        self.nuclide = "".join(line[4:11].split()).strip()
        self.reaction = line[11:21].strip()
        self.daughter = "".join(line[22:30].split()).strip()
        self.xs = float(line[29:41])
        self.delta_xs = float(line[44:56])
        return self

    def __repr__(self):
        daughter = "None"
        if self.daughter:
            daughter = self.daughter
        return "{:5} -> {:5}, {:10} @ {:.5e} +- {:.5e}".format(self.nuclide, 
                        daughter, self.reaction, self.xs, self.delta_xs)

@freeze_it
class PrintLib4(JSONSerializable):
    """
        An object to represent the Printlib4 output
    """
    def __init__(self):
        self.cross_sections = []

    def __len__(self):
        return len(self.cross_sections)

    def __getitem__(self, index):
        return self.cross_sections[index]

    def json_deserialize(self, j, objtype=object):
        super().json_deserialize(j)
        self.json_deserialize_list(j, 'cross_sections', CollapsedXSData)

    def fispact_deserialize(self, filerecord):
        self.__init__()

        for i in range(filerecord.start_index, filerecord.end_index):
            line = filerecord.cachedlines[i].strip('\n')
            self.cross_sections.append(CollapsedXSData().fispact_deserialize(line[:60]))
            self.cross_sections.append(CollapsedXSData().fispact_deserialize(line[61:]))


class PrintLib4Reader(Reader):
    """
        It can read fispact printlib 4 file formats
    """
    def __init__(self, filename):
        super().__init__(filename)
        self.record = PrintLib4FileRecord(filename)
        self.output = PrintLib4()

    def __enter__(self):
        self.output.fispact_deserialize(self.record)
        return self.output


import json

from pypact.util.decorators import freeze_it
from pypact.util.jsonserializable import JSONSerializable
from pypact.output.doserate import DoseRate
from pypact.output.nuclides import Nuclides
from pypact.output.nuclide import Nuclide
from pypact.output.gammaspectrum import GammaSpectrum
from pypact.output.tags import TIME_STEP_HEADER
import pypact.util.propertyfinder as pf

TIME_STEP_IGNORES = []


@freeze_it
class TimeStep(JSONSerializable):


    def __init__(self, ignorenuclides=False):
        self.irradiation_time = 0.0
        self.cooling_time = 0.0
        self.duration = 0.0
        self.flux = 0.0
        self.total_heat = 0.0
        self.alpha_heat = 0.0
        self.beta_heat = 0.0
        self.gamma_heat = 0.0
        self.ingestion_dose = 0.0
        self.inhalation_dose = 0.0
        self.total_activity = 0.0
        self.total_activity_exclude_trit = 0.0
        self.alpha_activity = 0.0
        self.beta_activity = 0.0
        self.gamma_activity = 0.0
        self.initial_mass = 0.0
        self.total_mass = 0.0
        self.number_of_fissions = 0.0
        self.burnup = 0.0
        self.gamma_spectrum = GammaSpectrum()
        self.total_displacement_rate = 0.0
        self.dose_rate = DoseRate()
        self.nuclides = []
        self.__ignorenuclides = ignorenuclides

    @property
    def isirradiation(self):
        return self.cooling_time == 0.0

    @property
    def currenttime(self):
        if self.isirradiation:
            return self.irradiation_time
        return self.cooling_time

    def fispact_deserialize(self, filerecord, interval):

        # reset to defaults before reading
        self.__init__(ignorenuclides=self.__ignorenuclides)

        substring = filerecord[interval]

        if substring:
            self.irradiation_time = filerecord.cumulirradiationtime(interval)
            self.cooling_time = filerecord.cumulcoolingtime(interval)

            def get_value(starttag, endtag):
                return pf.first(datadump=substring,
                                headertag=TIME_STEP_HEADER,
                                starttag=starttag,
                                endtag=endtag,
                                ignores=TIME_STEP_IGNORES,
                                asstring=False)

            self.flux = get_value(starttag='* * * FLUX AMP IS', endtag='/cm^2/s')

            self.alpha_heat = get_value(starttag='TOTAL ALPHA HEAT PRODUCTION', endtag='kW')
            self.beta_heat = get_value(starttag='TOTAL BETA  HEAT PRODUCTION', endtag='kW')
            self.gamma_heat = get_value(starttag='TOTAL GAMMA HEAT PRODUCTION', endtag='kW')
            self.total_heat = self.alpha_heat + self.beta_heat + self.gamma_heat

            self.initial_mass = get_value(starttag='0  INITIAL TOTAL MASS OF MATERIAL', endtag='kg')
            self.total_mass = get_value(starttag='0  TOTAL MASS OF MATERIAL', endtag='kg')

            self.number_of_fissions = get_value(starttag='NUMBER OF FISSIONS', endtag='BURN-UP')
            self.burnup = get_value(starttag='BURN-UP OF ACTINIDES', endtag='%')

            self.ingestion_dose = get_value(starttag='INGESTION  HAZARD FOR ALL MATERIALS', endtag='Sv/kg')
            self.inhalation_dose = get_value(starttag='INHALATION HAZARD FOR ALL MATERIALS', endtag='Sv/kg')

            self.alpha_activity = get_value(starttag='ALPHA BECQUERELS =', endtag='BETA')
            self.beta_activity = get_value(starttag='BETA BECQUERELS =', endtag='GAMMA')
            self.gamma_activity = get_value(starttag='GAMMA BECQUERELS =', endtag='')

            self.total_activity = get_value(starttag='TOTAL ACTIVITY FOR ALL MATERIALS', endtag='Bq')
            self.total_activity_exclude_trit = get_value(starttag='TOTAL ACTIVITY EXCLUDING TRITIUM', endtag='Bq')

            self.dose_rate.fispact_deserialize(filerecord, interval)
            self.gamma_spectrum.fispact_deserialize(filerecord, interval)

            self.total_displacement_rate = pf.first(
                datadump=substring,
                headertag="Total Displacement Rate (n,Dtot ) =",
                starttag="Displacements/sec  =",
                endtag="Displacements Per Atom/sec  =",
                ignores=TIME_STEP_IGNORES,
                asstring=False
            )
            self.duration = filerecord.times[interval - 1]


            # for fission this can be very large and hence very slow
            # if you do not care about the nuclides then we can set
            # it to ignore nuclides
            if not self.__ignorenuclides:
                t = Nuclides()
                t.fispact_deserialize(filerecord, interval)
                self.nuclides = t.nuclides


    def json_deserialize(self, json_dump, objtype=object):

        self.__init__(ignorenuclides=self.__ignorenuclides)

        d = JSONSerializable.json_deserialize(self, json_dump, objtype)

        if not self.__ignorenuclides:
            key = 'nuclides'
            if key in d:
                self.json_deserialize_list(json.dumps(d[key]), key, objtype=Nuclide)



import re

from pypact.output.tags import GAMMA_SPECTRUM_SUB_HEADER
from pypact.util.decorators import freeze_it
from pypact.util.jsonserializable import JSONSerializable

FLOAT_NUMBER = r"[0-9]+(?:\.(?:[0-9]+))?(?:e?(?:[-+]?[0-9]+)?)?"
GAMMA_SPECTRUM_LINE = \
    r"[^(]*\(\s*(?P<lb>{FN})\s*-\s*(?P<ub>{FN})\s*MeV\)\s*(?P<value>{FN}).*".format(
        FN=FLOAT_NUMBER,
    )
GAMMA_SPECTRUM_LINE_MATCHER = re.compile(GAMMA_SPECTRUM_LINE, re.IGNORECASE)


@freeze_it
class GammaSpectrum(JSONSerializable):
    """
        The gamma spectrum type from the output
    """

    def __init__(self):
        self.boundaries = []  # TODO dvp: should be numpy arrays (or even better xarrays)
        self.values = []

    def fispact_deserialize(self, file_record, interval):
        self.__init__()

        lines = file_record[interval]

        def extract_boundaries_and_values(_lines):
            header_found = False
            for line in _lines:
                if not header_found:
                    if GAMMA_SPECTRUM_SUB_HEADER in line:
                        header_found = True
                if header_found:
                    if line.strip() == "":
                        return
                    match = GAMMA_SPECTRUM_LINE_MATCHER.match(line)
                    lower_boundary = float(match.group("lb"))
                    upper_boundary = float(match.group("ub"))
                    value = float(match.group("value"))
                    yield lower_boundary, upper_boundary, value

        boundaries = []
        values = []

        for lb, ub, v in extract_boundaries_and_values(lines):
            if not boundaries:
                boundaries.append(lb)
            boundaries.append(ub)
            values.append(v)

        if values:
            self.boundaries = boundaries
            self.values = values



from setuptools import setup


setup(name='pypact',
      version='1.3.2',
      description='The package for reading and manipulating the fispact output text file.',
      url='https://github.com/fispact/pypact',
      author='UKAEA',
      author_email='thomas.stainer@ukaea.uk',
      license='Apache License 2.0',
      packages=[
            'pypact',
            'pypact.analysis',
            'pypact.library',
            'pypact.input',
            'pypact.output',
            'pypact.printlib',
            'pypact.tools',
            'pypact.util',
      ],
      install_requires=[],
      python_requires='>=3',
      scripts=[
           'pypact/tools/fispactconverter.py'
      ],
      setup_requires=[
            'pytest-runner',
      ],
      test_suite='tests.testsuite',
      tests_require=[
            'pytest',
            'pytest-cov>=2.3.1',
            'mock',
            'jsonschema',
            'numpy',
      ],
      zip_safe=False,
      entry_points={
            'console_scripts': [
                  'fispactconverter = pypact.tools.fispactconverter:main',
            ]
      },
      )
